"""ATTENTION:
This file is generated by exasol/toolbox/nox/_package_version.py when using:
   * either "poetry run -- nox -s project:fix"
   * or "poetry run -- nox -s version:check -- --fix"
Do not edit this file manually!
If you need to change the version, do so in the pyproject.toml, e.g. by using
`poetry version X.Y.Z`.
"""

MAJOR = 1
MINOR = 1
PATCH = 0
VERSION = f"{MAJOR}.{MINOR}.{PATCH}"
__version__ = VERSION
